@@ -32,10 +32,10 @@ inline void glmmr::MatrixW<modeltype>::update(){
   ArrayXd nvar_par(model.n());
   ArrayXd xb(model.n());
   switch(model.family.family){
-  case Fam::gaussian:
+  case Fam::gaussian: 
     nvar_par = model.data.variance;
     break;
-  case Fam::gamma:
+  case Fam::gamma: 
     nvar_par = model.data.variance.inverse();
     break;
   case Fam::beta:
@@ -44,8 +44,16 @@ inline void glmmr::MatrixW<modeltype>::update(){
   case Fam::binomial:
     nvar_par = model.data.variance.inverse();
     break;
+  case Fam::quantile: case Fam::quantile_scaled:
+    {
+      double qvar = (1 - 2*model.family.quantile + 2*model.family.quantile*model.family.quantile)/(model.family.quantile*model.family.quantile*(1-model.family.quantile)*(1-model.family.quantile));
+      if(model.family.family == Fam::quantile_scaled) qvar *= model.data.var_par*model.data.var_par;
+      nvar_par.setConstant(qvar);
+      break;
+    }
   default:
     nvar_par.setConstant(1.0);
+    break;
   }
   
   if(attenuated){
